package cli

import (
	"context"
	"encoding/json"
	"io"
	"log"
	"os"
	"os/signal"
	"sync"
	"syscall"

	"github.com/it-novum/event-collectd/gearman"
	"github.com/it-novum/event-collectd/icinga"
	"github.com/it-novum/event-collectd/message"
	"github.com/it-novum/event-collectd/oitc"
)

type CliArguments struct {
	Verbose *bool //Enable verbose logging
}

type EventCollectd struct {
	fifo          message.SyncMessageFIFO    // First in first out cache holding check results and other commands that needs to be passed to the gearman queue
	localGearman  *gearman.Client            // Gearman Client connected to local gearman job server
	dbWorker      oitc.DBWorker              // Database thread to find downtimes
	icingaPollers []*icinga.IcingaHttpClient // All running icinga pollers
}

type IniConfig struct {
	Gearman string `ini:"gearman"`
	Proxy   string `ini:"proxy"`
	LogFile string `ini:"log-file"`
	MyCNF   string `ini:"mycnf"`
}

// closeIcingaPollers is a helper that closes all icinga HTTP pollers
func closeIcingaPollers(icingaPollers []*icinga.IcingaHttpClient) {
	if len(icingaPollers) > 0 {
		var wg sync.WaitGroup
		wg.Add(len(icingaPollers))
		for i, icingaPoller := range icingaPollers {
			log.Println("Close Icinga2 Poller #", i)
			go func(icingaPoller *icinga.IcingaHttpClient) {
				icingaPoller.Close()
				wg.Done()
			}(icingaPoller)
		}
		wg.Wait()
	}
}

// MainThreadLoop is the main Thread of the Program
// It binds the signal handler and waits for signals such as SIGTERM to exit
// or SIGHUP to perform an reload
func MainThreadLoop(args *CliArguments) {
	// Bind signal handler for linux kernel signals
	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM, syscall.SIGHUP)

	shutdown := make(chan struct{})

	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	// Copy value of verbose
	// args.Verbose is a *bool but we want a bool (no pointer just the value)
	verbose := *args.Verbose

	// Main loop of the main thread
	// Basically this does nothing and just wait to receive a signal to stop (like SIGTERM)
	for {

		log.Println("Start event-collectd")
		// Start the "Run()" function in a separate thread
		go Run(ctx, shutdown, verbose)

		select {
		case <-ctx.Done():
			return

		case sig := <-sigs:
			if sig.String() == "hangup" {
				log.Println("Start reload of event-collectd")

				// Stop the Run() function but do not exit the program.
				// This will trigger a reload because the outer for loop will call the Run() function again.
				shutdown <- struct{}{}
			} else {
				// Stop Process by returning from this function (MainThreadLoop)
				log.Printf("Catch signal: %v - %v", sig, sig.String())
				return
			}
		}
	}
}

// Does the actual work of the event-collectd
func Run(pctx context.Context, shutdown chan struct{}, verbose bool) {
	// Todo add config file parsing first
	cfg := &IniConfig{
		Gearman: "127.0.0.1:4730",
		Proxy:   "",
		LogFile: "-",
		MyCNF:   "/opt/openitc/etc/mysql/mysql.cnf",
	}

	var err error

	// Setup logging
	logWriter := io.Discard
	if cfg.LogFile != "" {
		if cfg.LogFile == "-" {
			logWriter = os.Stdout
		} else {
			logWriter, err = os.OpenFile(cfg.LogFile, os.O_APPEND|os.O_CREATE, os.FileMode(0o644))
			if err != nil {
				log.Println(err)
				os.Exit(1)
			}
		}
	}
	log.SetOutput(logWriter)

	ctx, cancel := context.WithCancel(pctx)

	// Create main app and connect to the local gearman server
	app := &EventCollectd{
		localGearman: &gearman.Client{
			ClientOptions: gearman.ClientOptions{
				ClientName: "Local",
				Address:    cfg.Gearman,
			},
		},
	}

	// Start local FIFO where all threads can store check results
	// which should be send to local gearman server / openITCOCKPIT
	app.fifo.Start()

	// Connect to the openITCOCKPIT MySQL database
	dbh, err := oitc.NewDatabase(ctx, cfg.MyCNF)
	if err != nil {
		log.Print(err)
	}

	externalMonitorings, err := dbh.GetExternalSystems()
	if err != nil {
		log.Print(err)
	}

	// Start new database worker thread
	app.dbWorker.Start(ctx, dbh, app.fifo.DirectRawCommand())

	// Create mapping of host names and service descriptions to corresponding UUIDs
	log.Println("Create mapping of host and service names")
	mapping := oitc.NewOitcMapping(dbh)
	mapping.Reload()

	// This will start the gearman client in a separate thread
	app.localGearman.Start(ctx, &app.fifo)

	// Go lang will execute the last defined "defer" first. To keep it simple we try to avoid having more than one "defer" in out function
	defer func() {
		cancel()

		closeIcingaPollers(app.icingaPollers)

		// Keep the order
		app.dbWorker.Stop()
		app.fifo.Close()
		app.localGearman.Stop()
		dbh.Close()
	}()

	// Start two Icinga Pollers to simulate multiple instance
	for i, externalMonitoring := range externalMonitorings {
		switch externalMonitoring.SystemType {

		case "icinga2":
			var icingaConfig icinga.Config
			err = json.Unmarshal([]byte(externalMonitoring.JsonData), &icingaConfig)
			if err != nil {
				log.Println("Error while parsing JSON config for Icinga2")
				log.Println(externalMonitoring.JsonData)
				continue
			}

			icingaConfig.Verbose = verbose
			icingaPoller, err := icinga.New(icingaConfig)
			if err != nil {
				log.Print(err)
			}

			// Get objects to subscribe to StateChange (and other) events
			options := icinga.GetOptions(icingaConfig, i)

			// Start icinga poller in new Thread (go func)
			icingaPoller.Start(ctx, options, app.fifo.DirectCheckResult(), app.fifo.DirectRawCommand(), app.dbWorker.DirectCancelDowntime(), mapping)

			// Store Icinga Poller to local hashmap
			app.icingaPollers = append(app.icingaPollers, icingaPoller)

		default:
			log.Printf("System type '%v' is not supported yet.", externalMonitoring.SystemType)
		}

	}

	// Just wait for a shutdown signal
	// All the work is done by the go routines started above
	for {
		select {
		case <-ctx.Done():
			return

		case <-shutdown:
			return
		}
	}
}
