package gearman

import (
	"context"
	"log"
	"sync"
	"time"

	"github.com/it-novum/event-collectd/message"
	"github.com/it-novum/event-collectd/oitc"
	"github.com/it-novum/gearman-go/client"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

// ConnectionStatus is an enum holding the current status of the connection
type ConnectionStatus int

const (
	// Connected is returned if the client has an active connection to the server
	Connected ConnectionStatus = 0
	// Disconnected is returned if the client can't connect to the server
	Disconnected ConnectionStatus = 1
	// Stopped is returned if the client has been closed
	Stopped ConnectionStatus = 2
)

type ClientOptions struct {
	ClientName string
	Address    string
}

// Client represents the connection to gearman server
type Client struct {
	ClientOptions

	mtx      sync.Mutex
	wg       sync.WaitGroup
	shutdown chan struct{}

	fifo   *message.SyncMessageFIFO
	client *client.Client

	statusClient ConnectionStatus
}

// IsRunning returns true if gearman client is running
func (c *Client) IsRunning() bool {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if c.client != nil && c.statusClient == Connected {
		return true
	}
	return false
}

// Set client status to Disconnected in case of any error
func (c *Client) handleClientError(err error) {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if c.statusClient == Connected {
		log.Println(&Error{clientName: c.ClientName, err: err, msg: "connection failed"})
		c.statusClient = Disconnected
		if c.client != nil {
			c.client.Close()
		}
		c.client = nil
	}
}

// Create new GearmanClient to put messages into a gearman queue
func (c *Client) connectClient() {
	c.mtx.Lock()
	defer c.mtx.Unlock()

	if c.client == nil {
		log.Println(&Error{clientName: c.ClientName, msg: "connect to gearman server as client"})
		gc, err := client.New("tcp", c.Address)
		if err != nil {
			log.Println(&Error{clientName: c.ClientName, err: err, msg: "connection failed"})
			return
		}

		gc.ErrorHandler = c.handleClientError

		c.client = gc
		c.statusClient = Connected
	} else if c.statusClient == Disconnected {
		c.statusClient = Connected
	}
}

// run the event loop in a go func in background
func (c *Client) run(parent context.Context) {
	ctx, cancel := context.WithCancel(parent)
	defer cancel()

	defer c.wg.Done()

	defer func() {
		if c.client != nil {
			log.Println(&Error{clientName: c.ClientName, msg: "close gearman client connection"})
			_ = c.client.Close()
			c.client = nil
		}
		log.Println(&Error{clientName: c.ClientName, msg: "close gearman fifo"})
		c.statusClient = Stopped
	}()

	// Check all 5 seconds if connection is established and reconnect if necessary
	ticker := time.NewTicker(time.Second * 5)
	defer ticker.Stop()

	// Init connect on startup
	c.connectClient()

	// Create a fake channel
	// If we are not connected to the gearman job server we do not want to receive the "notify" signal from FIFO
	// Otherwise we would lose check results into the void
	noNotify := make(chan struct{})

	for {
		var notify chan struct{}
		if c.IsRunning() {
			// Only listen to notify signal from FIFO if gearman connection is established
			notify = c.fifo.Notify()
		} else {
			log.Println(&Error{clientName: c.ClientName, msg: "Waiting for gearman reconnect"})
			notify = noNotify
		}

		select {
		case <-ctx.Done():
			return
		case _, more := <-c.shutdown:
			if !more {
				// Channel c.shutdown got closed - drop gearman client
				return
			}
		case <-ticker.C:
			// Check if we are still connected to the gearman job server
			c.connectClient()
		//case <-c.fifo.Notify():
		case <-notify:
			// Got notify signal from FIFO
			// Pass messages from the FIFO cache to the gearman job server
			// But first make sure that we are still connected
			//fmt.Println("GOT NOTIFY FROM FIFO")
			if c.IsRunning() {
				checkResults := c.fifo.GetAllCheckResults(50)

				// Refactor this to Bulk Messages some day
				// https://wiki.it-novum.com/de/openITCOCKPIT-v4/47088223
				// Send check results to the Statusengine Broker (via Gearman)
				for _, cr := range checkResults {
					bcr := oitc.BrokerCheckResult{
						Command: "check_result",
						Data:    cr,
					}

					json, err := json.Marshal(bcr)
					if err != nil {
						log.Println("Error while encoding check results as JSON")
						log.Println(err)
						continue
					}

					if err := c.SendRawMessage("statusngin_cmd", string(json)); err != nil {
						c.handleClientError(err)
						break
					}
				}

				// Send external commands to the Statusengine Broker (via Gearman)
				rawCommands := c.fifo.GetRawCommands(50)
				for _, externalCommand := range rawCommands {

					json, err := json.Marshal(externalCommand)
					if err != nil {
						log.Println("Error while encoding external command as JSON")
						log.Println(err)
						continue
					}

					if err := c.SendRawMessage("statusngin_cmd", string(json)); err != nil {
						c.handleClientError(err)
						break
					}
				}
			}
		}
	}
}

// Start the event loop and returns if successful
func (c *Client) Start(ctx context.Context, fifo *message.SyncMessageFIFO) {
	c.wg.Add(1)

	// Signal this channel to stop the gearman client
	c.shutdown = make(chan struct{})

	c.fifo = fifo

	// Gearman Client is not connected yet
	c.statusClient = Disconnected

	// Start new gearman client in separate go func (thread) and return from this function
	go c.run(ctx)
}

// Stop should be called before programs exit
func (c *Client) Stop() {
	close(c.shutdown)
	c.wg.Wait()
}

func (c *Client) SendRawMessage(queue, data string) error {
	_, err := c.client.DoBg(queue, []byte(data), client.JobNormal)
	if err != nil {
		return &Error{clientName: c.ClientName, err: err, msg: "could not send gearman worker job to server"}
	}
	return nil
}
