package icinga

import (
	"bufio"
	"bytes"
	"context"
	"crypto/tls"
	"fmt"
	"log"
	"net"
	"net/http"
	"net/url"
	"strings"
	"sync"
	"time"

	"github.com/it-novum/event-collectd/helper"
	"github.com/it-novum/event-collectd/oitc"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

type IcingaHttpClient struct {
	ctx                   context.Context
	config                Config
	client                http.Client
	checkResultChannel    chan *oitc.CheckResult
	rawCommandChannel     chan *oitc.BrokerRawCommand
	cancelDowntimeChannel chan *oitc.CancelDowntime
	mapping               *oitc.OpenITCOCKPITMapping
	options               Register
	wg                    sync.WaitGroup
	shutdown              chan struct{}
}

type Register struct {
	Queue string   `json:"queue"`
	Types []string `json:"types"`
}

// Returns an "instance" of the Register struct which contains all required configuration
// options to subscribe to the Icinga Event Stream API
func GetOptions(config Config, pollerIndex int) Register {

	types := []string{"CheckResult"}
	if config.CheckType == "StateChange" {
		// Only subscribe to state change events
		types = []string{"StateChange"}
	}

	if config.ReceiveAcknowledgements == 1 {
		types = append(types, "AcknowledgementSet", "AcknowledgementCleared")
	}

	if config.ReceiveDowntimes == 1 {
		types = append(types, "DowntimeAdded", "DowntimeRemoved", "DowntimeStarted")
	}

	return Register{
		Queue: fmt.Sprintf("go-oitc-%v", pollerIndex),
		Types: types,
	}
}

// New creates a HTTP client to query the Icinga2 Event Stream API
// https://icinga.com/docs/icinga-2/latest/doc/12-icinga2-api/#icinga2-api-event-streams
// Plain curl example
// curl -k -s -u root:password -H 'Accept: application/json' -X POST 'https://icinga2.itsm.love:5665/v1/events?queue=m&types=CheckResult'
func New(config Config) (*IcingaHttpClient, error) {
	var proxyFunc func(*http.Request) (*url.URL, error) = nil
	if config.ProxyAddress != "" {
		proxyURL, err := url.Parse(config.ProxyAddress)
		if err != nil {
			return nil, err
		}
		proxyFunc = http.ProxyURL(proxyURL)
	}

	return &IcingaHttpClient{
		config: config,
		client: http.Client{
			Timeout: 0, //5 * time.Minute,
			Transport: &http.Transport{
				TLSHandshakeTimeout: 10 * time.Second,
				TLSClientConfig: &tls.Config{
					InsecureSkipVerify: config.IgnoreSslCertificate == 1,
				},
				Proxy: proxyFunc,
				DialContext: (&net.Dialer{
					Timeout:   30 * time.Second,
					KeepAlive: 30 * time.Second,
				}).DialContext,
			},
		},
	}, nil
}

func (c *IcingaHttpClient) Close() {
	close(c.shutdown)
	c.wg.Wait()
}

func (c *IcingaHttpClient) connectEventStream() error {
	defer c.wg.Done()
	for {

		select {
		case <-c.ctx.Done():
			return nil
		default:
		}

		log.Println("Connecting to Icinga Event Stream API")
		err := c.readEventStream()
		if err == context.Canceled {
			log.Println("Context canceled. Shutdown Icinga Event Stream")
			return nil
		}

		if err != nil {
			log.Println("Trying to reconnect to Icinga Event Stream API in 5 seconds")
			time.Sleep(5 * time.Second) // Do not DDOS the Icinga API
		}
	}
}

func (c *IcingaHttpClient) readEventStream() error {
	requestOptions, _ := json.Marshal(c.options)

	// URL of the Icinga Event Stream API E.g.: "https://icinga2.itsm.love:5665/v1/events"
	url := c.config.ApiUrl + "/v1/events"

	request, err := http.NewRequestWithContext(c.ctx, "POST", url, bytes.NewBuffer(requestOptions))
	if err != nil {
		return err
	}

	request.Header.Set("Accept", "application/json")
	request.SetBasicAuth(c.config.ApiUser, c.config.ApiPassword)

	response, err := c.client.Do(request)
	if err != nil {
		log.Println("Server error: ", err)
		return err
	}
	defer response.Body.Close()

	// Read response and send it to channel
	reader := bufio.NewReader(response.Body)

	// Go to endless reading loop
	for {
		select {
		case <-c.ctx.Done():
			return nil
		default:
		}

		line, err := reader.ReadBytes('\n')

		if err != nil {
			log.Println("Error reading stream: ", err)
			return err
		}

		//fmt.Println(string(line))

		var eventType EventType
		_ = json.Unmarshal(line, &eventType)

		// Handle received events
		switch eventType.Type {
		case "CheckResult":
			var icingaCheckResult EventTypeCheckResult
			err = json.Unmarshal(line, &icingaCheckResult)
			if err != nil {
				log.Println("Error while parsing JSON data from Icinga")
				log.Println(string(line))
				continue
			}

			// Sometimes Icinga sends "Nagios like" performance data as string arrays
			// Like so:[ "rta=0.048000ms;100.000000;500.000000;0.000000", "pl=0%;20;60;0" ]
			// Sometimes it sends a object with already processed and parsed performance data
			// Like so [ { "counter":false,"crit":null,"label":"api_num_conn_endpoints","max":null,"min":null,"type":"PerfdataValue","unit":"","value":2,"warn":null } ]
			// There is probably a better solution to detect this - if so - feel free to refactor this
			var icingaPerfdataStringArray []string
			var icingaPerfdataObjectArray []PerformanceDataValue
			perfdataStringErr := json.Unmarshal([]byte(icingaCheckResult.CheckResult.PerformanceDataInterface.ToString()), &icingaPerfdataStringArray)
			if perfdataStringErr != nil {
				icingaPerfdataStringArray = []string{}
				// Maybe performance data this is an object?
				perfdataObjectErr := json.Unmarshal([]byte(icingaCheckResult.CheckResult.PerformanceDataInterface.ToString()), &icingaPerfdataObjectArray)
				if perfdataObjectErr != nil {
					icingaPerfdataObjectArray = []PerformanceDataValue{}
					if c.config.Verbose {
						log.Println("Error while parsing Icinga performance data JSON. Data was not a valid array of strings")
						log.Println(perfdataStringErr)

						log.Println("Error while parsing Icinga performance data JSON. Data was not a valid object of arrays")
						log.Println(perfdataObjectErr)
					}
				}
			}

			checkType := oitc.HOST_CHECK
			if icingaCheckResult.Service != "" {
				checkType = oitc.SERVICE_CHECK
			}

			// https://icinga.com/docs/icinga-2/latest/doc/08-advanced-topics/#advanced-value-types-checkresult
			rc, _ := icingaCheckResult.CheckResult.State.Int64()

			hostUuid, err := c.mapping.GetHostUuidByHostname(icingaCheckResult.Host)
			if err != nil {
				if c.config.Verbose {
					log.Println(err)
				}
				continue
			}

			if checkType == oitc.HOST_CHECK {
				rc, err = oitc.InHostcheckBoundaries(convertIcingaHoststatusToNaemon(rc), icingaCheckResult.Host)
				if err != nil {
					log.Println(err)
				}

				// Send host check result to FIFO
				c.checkResultChannel <- &oitc.CheckResult{
					HostName:           hostUuid,
					ServiceDescription: "",
					Output:             icingaCheckResult.CheckResult.Output,
					LongOutput:         "", // Not available in Icinga
					PerfData:           getPerformanceDataAsString(icingaPerfdataStringArray, icingaPerfdataObjectArray),
					CheckType:          int64(checkType),
					ReturnCode:         rc,
					StartTime:          int64(icingaCheckResult.CheckResult.ExecutionStart),
					EndTime:            int64(icingaCheckResult.CheckResult.ExecutionEnd),
					EarlyTimeout:       0,
					Latency:            0.0,
					ExitedOk:           1,
				}
			} else {
				// Process service check result
				serviceUuid, err := c.mapping.GetServiceUuidByHostUuidAndServiceDescription(hostUuid, icingaCheckResult.Service)
				if err != nil {
					if c.config.Verbose {
						log.Println(err)
					}
					continue
				}

				rc, err = oitc.InServicecheckBoundaries(rc, icingaCheckResult.Host, icingaCheckResult.Service)
				if err != nil {
					log.Println(err)
				}

				// Send service check result to FIFO
				c.checkResultChannel <- &oitc.CheckResult{
					HostName:           hostUuid,
					ServiceDescription: serviceUuid,
					Output:             icingaCheckResult.CheckResult.Output,
					LongOutput:         "", // Not available in Icinga
					PerfData:           getPerformanceDataAsString(icingaPerfdataStringArray, icingaPerfdataObjectArray),
					CheckType:          int64(checkType),
					ReturnCode:         rc,
					StartTime:          int64(icingaCheckResult.CheckResult.ExecutionStart),
					EndTime:            int64(icingaCheckResult.CheckResult.ExecutionEnd),
					EarlyTimeout:       0,
					Latency:            0.0,
					ExitedOk:           1,
				}
			}

		case "DowntimeAdded":
			var icingaDowntimeAdd EventDowntimeAdded
			err = json.Unmarshal(line, &icingaDowntimeAdd)
			if err != nil {
				log.Println("Error while parsing JSON data from Icinga")
				log.Println(string(line))
				continue
			}

			hostUuid, err := c.mapping.GetHostUuidByHostname(icingaDowntimeAdd.Downtime.HostName)
			if err != nil {
				if c.config.Verbose {
					log.Println(err)
				}
				continue
			}

			startTime := helper.JsonToInt64(icingaDowntimeAdd.Downtime.StartTime)
			endTime := helper.JsonToInt64(icingaDowntimeAdd.Downtime.EndTime)

			if icingaDowntimeAdd.Downtime.ServiceName == "" {
				// Create host downtime
				cmd := fmt.Sprintf(
					"[%d] SCHEDULE_HOST_DOWNTIME;%s;%d;%d;%d;%d;%d;%s;%s",
					time.Now().Unix(),
					hostUuid,
					startTime,
					endTime,
					1, // fixed
					0, // trigger_id
					icingaDowntimeAdd.Downtime.Duration,
					icingaDowntimeAdd.Downtime.Author,
					icingaDowntimeAdd.Downtime.Comment,
				)

				//  Send downtime as external command to FIFO
				c.rawCommandChannel <- &oitc.BrokerRawCommand{
					Command: "raw",
					Data:    cmd,
				}
			} else {
				// Create service downtime

				serviceUuid, err := c.mapping.GetServiceUuidByHostUuidAndServiceDescription(hostUuid, icingaDowntimeAdd.Downtime.ServiceName)
				if err != nil {
					if c.config.Verbose {
						log.Println(err)
					}
					continue
				}

				cmd := fmt.Sprintf(
					"[%d] SCHEDULE_SVC_DOWNTIME;%s;%s;%d;%d;%d;%d;%d;%s;%s",
					time.Now().Unix(),
					hostUuid,
					serviceUuid,
					startTime,
					endTime,
					1, // fixed,
					0, // trigger_id
					icingaDowntimeAdd.Downtime.Duration,
					icingaDowntimeAdd.Downtime.Author,
					icingaDowntimeAdd.Downtime.Comment,
				)

				//  Send downtime as external command to FIFO
				c.rawCommandChannel <- &oitc.BrokerRawCommand{
					Command: "raw",
					Data:    cmd,
				}
			}

		case "DowntimeRemoved":
			var icingaDowntimeRemoved EventDowntimeRemoved
			err = json.Unmarshal(line, &icingaDowntimeRemoved)
			if err != nil {
				log.Println("Error while parsing JSON data from Icinga")
				log.Println(string(line))
				continue
			}

			removeTime := helper.JsonToInt64(icingaDowntimeRemoved.Downtime.RemoveTime)
			// https://github.com/Icinga/icinga2/blob/66b039df9c5460bcd4db4c4774e09a1ba8ca075b/lib/icinga/downtime.ti#L72-L74
			// https://github.com/Icinga/icinga2/blob/66b039df9c5460bcd4db4c4774e09a1ba8ca075b/lib/db_ido/dbevents.cpp#L621
			wasCancelled := removeTime > 0

			var serviceUuid string
			if wasCancelled == true {
				hostUuid, err := c.mapping.GetHostUuidByHostname(icingaDowntimeRemoved.Downtime.HostName)
				if err != nil {
					if c.config.Verbose {
						log.Println(err)
					}
					continue
				}

				if icingaDowntimeRemoved.Downtime.ServiceName != "" {
					serviceUuid, err = c.mapping.GetServiceUuidByHostUuidAndServiceDescription(hostUuid, icingaDowntimeRemoved.Downtime.ServiceName)
					if err != nil {
						if c.config.Verbose {
							log.Println(err)
						}
						continue
					}
				}

				startTime := helper.JsonToInt64(icingaDowntimeRemoved.Downtime.StartTime)
				endTime := helper.JsonToInt64(icingaDowntimeRemoved.Downtime.EndTime)

				//  Send downtime as external command to FIFO
				c.cancelDowntimeChannel <- &oitc.CancelDowntime{
					HostUuid:     hostUuid,
					ServiceUuid:  serviceUuid,
					Comment:      icingaDowntimeRemoved.Downtime.Comment,
					StartTime:    startTime,
					EndTime:      endTime,
					CanceledTime: removeTime,
				}

			}

		case "DowntimeStarted":
			// Not sure if we need to implement this at all
			//log.Println("TODO implement DowntimeStarted handling")

		case "AcknowledgementSet":
			var icingaAcknowledgementSet EventAcknowledgementSet
			err = json.Unmarshal(line, &icingaAcknowledgementSet)
			if err != nil {
				log.Println("Error while parsing JSON data from Icinga")
				log.Println(string(line))
				continue
			}

			hostUuid, err := c.mapping.GetHostUuidByHostname(icingaAcknowledgementSet.Host)
			if err != nil {
				if c.config.Verbose {
					log.Println(err)
				}
				continue
			}

			sticky := 0
			if v, _ := icingaAcknowledgementSet.AcknowledgementType.Int64(); v == 2 {
				sticky = 1
			}

			notify := 0
			if icingaAcknowledgementSet.Notify {
				notify = 1
			}

			if icingaAcknowledgementSet.Service == "" {
				// Create Host Acknowledgement
				cmd := fmt.Sprintf(
					"[%d] ACKNOWLEDGE_HOST_PROBLEM;%s;%d;%d;1;%s;%s",
					time.Now().Unix(),
					hostUuid,
					sticky,
					notify,
					icingaAcknowledgementSet.Author,
					icingaAcknowledgementSet.Comment,
				)

				//  Send acknowledgement as external command to FIFO
				c.rawCommandChannel <- &oitc.BrokerRawCommand{
					Command: "raw",
					Data:    cmd,
				}
			} else {
				// Create Service Acknowledgement

				serviceUuid, err := c.mapping.GetServiceUuidByHostUuidAndServiceDescription(hostUuid, icingaAcknowledgementSet.Service)
				if err != nil {
					if c.config.Verbose {
						log.Println(err)
					}
					continue
				}

				cmd := fmt.Sprintf(
					"[%d] ACKNOWLEDGE_SVC_PROBLEM;%s;%s;%d;%d;1;%s;%s",
					time.Now().Unix(),
					hostUuid,
					serviceUuid,
					sticky,
					notify,
					icingaAcknowledgementSet.Author,
					icingaAcknowledgementSet.Comment,
				)

				//  Send acknowledgement as external command to FIFO
				c.rawCommandChannel <- &oitc.BrokerRawCommand{
					Command: "raw",
					Data:    cmd,
				}
			}

		case "AcknowledgementCleared":
			var icingaAcknowledgementCleared EventAcknowledgementCleared
			err = json.Unmarshal(line, &icingaAcknowledgementCleared)
			if err != nil {
				log.Println("Error while parsing JSON data from Icinga")
				log.Println(string(line))
				continue
			}

			hostUuid, err := c.mapping.GetHostUuidByHostname(icingaAcknowledgementCleared.Host)
			if err != nil {
				if c.config.Verbose {
					log.Println(err)
				}
				continue
			}

			if helper.JsonToInt64(icingaAcknowledgementCleared.State) == 0 {
				// State change to Ok. openITCOCKPIT will remove the acknowledgement by itself
				continue
			}

			//acknowledgement got removed but the state is still not Ok

			if icingaAcknowledgementCleared.Service == "" {
				// Remove host acknowledgement
				cmd := fmt.Sprintf(
					"[%d] REMOVE_HOST_ACKNOWLEDGEMENT;%s",
					time.Now().Unix(),
					hostUuid,
				)

				//  Send remove acknowledgement as external command to FIFO
				c.rawCommandChannel <- &oitc.BrokerRawCommand{
					Command: "raw",
					Data:    cmd,
				}
			} else {
				// Remove service acknowledgement
				serviceUuid, err := c.mapping.GetServiceUuidByHostUuidAndServiceDescription(hostUuid, icingaAcknowledgementCleared.Service)
				if err != nil {
					if c.config.Verbose {
						log.Println(err)
					}
					continue
				}

				cmd := fmt.Sprintf(
					"[%d] REMOVE_SVC_ACKNOWLEDGEMENT;%s;%s",
					time.Now().Unix(),
					hostUuid,
					serviceUuid,
				)

				//  Send remove acknowledgement as external command to FIFO
				c.rawCommandChannel <- &oitc.BrokerRawCommand{
					Command: "raw",
					Data:    cmd,
				}

			}

		case "StateChange":
			var icingaStateChange EventTypeStateChange
			err = json.Unmarshal(line, &icingaStateChange)
			if err != nil {
				log.Println("Error while parsing JSON data from Icinga")
				log.Println(string(line))
				continue
			}

			// Sometimes Icinga sends "Nagios like" performance data as string arrays
			// Like so:[ "rta=0.048000ms;100.000000;500.000000;0.000000", "pl=0%;20;60;0" ]
			// Sometimes it sends a object with already processed and parsed performance data
			// Like so [ { "counter":false,"crit":null,"label":"api_num_conn_endpoints","max":null,"min":null,"type":"PerfdataValue","unit":"","value":2,"warn":null } ]
			// There is probably a better solution to detect this - if so - feel free to refactor this
			var icingaPerfdataStringArray []string
			var icingaPerfdataObjectArray []PerformanceDataValue
			perfdataStringErr := json.Unmarshal([]byte(icingaStateChange.CheckResult.PerformanceDataInterface.ToString()), &icingaPerfdataStringArray)
			if perfdataStringErr != nil {
				icingaPerfdataStringArray = []string{}
				// Maybe performance data this is an object?
				perfdataObjectErr := json.Unmarshal([]byte(icingaStateChange.CheckResult.PerformanceDataInterface.ToString()), &icingaPerfdataObjectArray)
				if perfdataObjectErr != nil {
					icingaPerfdataObjectArray = []PerformanceDataValue{}
					if c.config.Verbose {
						log.Println("Error while parsing Icinga performance data JSON. Data was not a valid array of strings")
						log.Println(perfdataStringErr)

						log.Println("Error while parsing Icinga performance data JSON. Data was not a valid object of arrays")
						log.Println(perfdataObjectErr)
					}
				}
			}

			checkType := oitc.HOST_CHECK
			if icingaStateChange.Service != "" {
				checkType = oitc.SERVICE_CHECK
			}

			// https://icinga.com/docs/icinga-2/latest/doc/08-advanced-topics/#advanced-value-types-checkresult
			rc, _ := icingaStateChange.CheckResult.State.Int64()

			hostUuid, err := c.mapping.GetHostUuidByHostname(icingaStateChange.Host)
			if err != nil {
				if c.config.Verbose {
					log.Println(err)
				}
				continue
			}

			if checkType == oitc.HOST_CHECK {
				rc, err = oitc.InHostcheckBoundaries(convertIcingaHoststatusToNaemon(rc), icingaStateChange.Host)
				if err != nil {
					log.Println(err)
				}

				// Send host state change to FIFO
				c.checkResultChannel <- &oitc.CheckResult{
					HostName:           hostUuid,
					ServiceDescription: "",
					Output:             icingaStateChange.CheckResult.Output,
					LongOutput:         "", // Not available in Icinga
					PerfData:           getPerformanceDataAsString(icingaPerfdataStringArray, icingaPerfdataObjectArray),
					CheckType:          int64(checkType),
					ReturnCode:         rc,
					StartTime:          int64(icingaStateChange.CheckResult.ExecutionStart),
					EndTime:            int64(icingaStateChange.CheckResult.ExecutionEnd),
					EarlyTimeout:       0,
					Latency:            0.0,
					ExitedOk:           1,
				}
			} else {
				// Process service state change event
				serviceUuid, err := c.mapping.GetServiceUuidByHostUuidAndServiceDescription(hostUuid, icingaStateChange.Service)
				if err != nil {
					if c.config.Verbose {
						log.Println(err)
					}
					continue
				}

				rc, err = oitc.InServicecheckBoundaries(rc, icingaStateChange.Host, icingaStateChange.Service)
				if err != nil {
					log.Println(err)
				}

				// Send service state change to FIFO
				c.checkResultChannel <- &oitc.CheckResult{
					HostName:           hostUuid,
					ServiceDescription: serviceUuid,
					Output:             icingaStateChange.CheckResult.Output,
					LongOutput:         "", // Not available in Icinga
					PerfData:           getPerformanceDataAsString(icingaPerfdataStringArray, icingaPerfdataObjectArray),
					CheckType:          int64(checkType),
					ReturnCode:         rc,
					StartTime:          int64(icingaStateChange.CheckResult.ExecutionStart),
					EndTime:            int64(icingaStateChange.CheckResult.ExecutionEnd),
					EarlyTimeout:       0,
					Latency:            0.0,
					ExitedOk:           1,
				}
			}

		default:
			log.Println("Received unknown event type from Icinga: ", eventType)
		}

	}
}

func getPerformanceDataAsString(icingaPerfdataStringArray []string, icingaPerfdataObjectArray []PerformanceDataValue) string {
	if len(icingaPerfdataObjectArray) > 0 {
		// Icinga return pre-processed performance data / performance data that was already parsed by Icinga
		// We need to re-write the performance data into something Naemon (and Nagios) can understand.

		icingaPerfdataStringArray = make([]string, len(icingaPerfdataObjectArray))
		for i, metric := range icingaPerfdataObjectArray {
			if metric.Counter {
				metric.Unit = "c"
			}
			perfdataString := fmt.Sprintf(
				"%v=%v%v;%v;%v;%v;%v",
				metric.Label,
				metric.Value.String(),
				metric.Unit,
				metric.Warn.String(),
				metric.Crit.String(),
				metric.Min.String(),
				metric.Max.String(),
			)

			icingaPerfdataStringArray[i] = perfdataString
		}
	}

	if len(icingaPerfdataStringArray) > 0 {
		// Icinga returned Nagios like performance data
		return strings.Join(icingaPerfdataStringArray[:], " ")
	}

	return ""
}

// Start reading from the Icinga Event Stream and returns immediately (SHOULD NOT RUN IN GOROUTINE)
func (c *IcingaHttpClient) Start(ctx context.Context, options Register, cr chan *oitc.CheckResult, raw chan *oitc.BrokerRawCommand, cancelDowntimeChannel chan *oitc.CancelDowntime, mapping *oitc.OpenITCOCKPITMapping) error {
	c.shutdown = make(chan struct{})
	c.checkResultChannel = cr
	c.rawCommandChannel = raw
	c.cancelDowntimeChannel = cancelDowntimeChannel
	c.mapping = mapping

	c.options = options

	c.wg.Add(1)
	go func() {
		ctx, cancel := context.WithCancel(ctx)
		c.ctx = ctx

		defer func() {
			cancel()
			c.wg.Done()
		}()

		// Send long running POST request to Icinga API in new thread (go func)
		c.wg.Add(1)
		go c.connectEventStream()

		for {
			select {
			case <-ctx.Done():
				return

			case _, ok := <-c.shutdown:
				if !ok {
					// Channel c.shutdown got closed - close this Icinga poller
					log.Println("Shutdown Icinga2 poller")
					return
				}
			}

		}

	}()

	return nil
}
