package message

import (
	"container/list"
	"log"
	"sync"

	"github.com/it-novum/event-collectd/oitc"
)

type EventChannels struct {
	Input chan *EventMessage
	Cr    chan *oitc.CheckResult
}

type EventMessage struct {
	Source  string
	Message string
}

type SyncMessageFIFO struct {
	mtx          sync.Mutex
	messages     list.List
	checkresults list.List
	rawCommands  list.List

	input  chan *EventMessage
	notify chan struct{}
	cr     chan *oitc.CheckResult
	raw    chan *oitc.BrokerRawCommand

	closer chan struct{}
	wg     sync.WaitGroup
}

func (f *SyncMessageFIFO) run() {
	defer func() {
		close(f.input)
		close(f.notify)
		close(f.cr)
		close(f.raw)
		f.wg.Done()
	}()

	for {

		//f.mtx.Lock() // This blocks all External Monitoring threads so do not do this!
		if f.checkresults.Len() > 0 {
			//fmt.Println("FIFO WILL NOTIFY CHECK RESULTS")
			// We have messages in cache - notify the Gearman Client about this
			f.notify <- struct{}{}
		}

		if f.rawCommands.Len() > 0 {
			//fmt.Println("FIFO WILL NOTIFY RAW COMMANDS")
			// We have messages in cache - notify the Gearman Client about this
			f.notify <- struct{}{}
		}
		//f.mtx.Unlock()

		select {
		case <-f.closer:
			log.Println("Shutdown FIFO")
			return

		case msg := <-f.input:
			// Got message on input channel (event from icinga or what ever)
			f.mtx.Lock()
			f.messages.PushBack(msg)
			//fmt.Println("FIFO got msg!")
			f.mtx.Unlock()

		case cr := <-f.cr:
			// Got check result on cr channel (event from icinga or what ever)
			f.mtx.Lock()
			f.checkresults.PushBack(cr)
			//fmt.Println("FIFO got check result!")
			//fmt.Println(cr)
			f.mtx.Unlock()

		case raw := <-f.raw:
			// Got raw naemon external command on raw channel (event from icinga or what ever)
			f.mtx.Lock()
			f.rawCommands.PushBack(raw)
			//fmt.Println("FIFO got raw command!")
			//fmt.Println(raw)
			f.mtx.Unlock()
		}
	}
}

// Start creates a background process for i/o
func (f *SyncMessageFIFO) Start() {
	f.input = make(chan *EventMessage)
	f.notify = make(chan struct{})
	f.cr = make(chan *oitc.CheckResult)
	f.raw = make(chan *oitc.BrokerRawCommand)
	f.closer = make(chan struct{})

	f.wg.Add(1)
	go f.run()
}

// Close message FIFO
func (f *SyncMessageFIFO) Close() {
	f.closer <- struct{}{}
	f.wg.Wait()
}

// GetAll returns up to max messages from the FIFO
func (f *SyncMessageFIFO) GetAll(max int) []*EventMessage {
	f.mtx.Lock()
	defer f.mtx.Unlock()

	lenMsg := f.messages.Len()
	if lenMsg > max {
		lenMsg = max
	}
	msgs := make([]*EventMessage, lenMsg)

	for i := 0; i < lenMsg; i++ {
		el := f.messages.Front()
		if el == nil {
			msgs = msgs[:i]
			break
		}
		msgs[i] = f.messages.Remove(el).(*EventMessage)
	}

	return msgs
}

// GetAllCheckResults returns up to max messages from the FIFO
func (f *SyncMessageFIFO) GetAllCheckResults(max int) []*oitc.CheckResult {
	f.mtx.Lock()
	defer f.mtx.Unlock()

	lenMsg := f.checkresults.Len()
	if lenMsg > max {
		lenMsg = max
	}
	msgs := make([]*oitc.CheckResult, lenMsg)

	for i := 0; i < lenMsg; i++ {
		el := f.checkresults.Front()
		if el == nil {
			msgs = msgs[:i]
			break
		}
		msgs[i] = f.checkresults.Remove(el).(*oitc.CheckResult)
	}

	return msgs
}

// GetRawCommands returns up to max messages from the FIFO
func (f *SyncMessageFIFO) GetRawCommands(max int) []*oitc.BrokerRawCommand {
	f.mtx.Lock()
	defer f.mtx.Unlock()

	lenMsg := f.rawCommands.Len()
	if lenMsg > max {
		lenMsg = max
	}
	msgs := make([]*oitc.BrokerRawCommand, lenMsg)

	for i := 0; i < lenMsg; i++ {
		el := f.rawCommands.Front()
		if el == nil {
			msgs = msgs[:i]
			break
		}
		msgs[i] = f.rawCommands.Remove(el).(*oitc.BrokerRawCommand)
	}

	return msgs
}

// Direct returns the channel were Icinga (or other sources) can push messages directly into the buffer
func (f *SyncMessageFIFO) Direct() chan *EventMessage {
	return f.input
}

// Notify returns the channel to trigger the gearman client that messages are waiting for processing
func (f *SyncMessageFIFO) Notify() chan struct{} {
	return f.notify
}

// DirectCheckResult returns the channel where Icinga (or other sources) can push check results directly into the buffer
func (f *SyncMessageFIFO) DirectCheckResult() chan *oitc.CheckResult {
	return f.cr
}

// DirectRawCommand returns the channel where Icinga (or other sources) can push raw external commands directly into the buffer
func (f *SyncMessageFIFO) DirectRawCommand() chan *oitc.BrokerRawCommand {
	return f.raw
}
