package oitc

import (
	"context"
	"database/sql"
	"fmt"
	"os"

	_ "github.com/go-sql-driver/mysql"
	"github.com/uptrace/bun"
	"github.com/uptrace/bun/dialect/mysqldialect"
	"gopkg.in/ini.v1"
)

const GENERIC_SERVICE = 1 << 0    //1
const EVK_SERVICE = 1 << 1        //2
const SLA_SERVICE = 1 << 2        //4
const MK_SERVICE = 1 << 3         //8
const OITC_AGENT_SERVICE = 1 << 4 //16
const PROMETHEUS_SERVICE = 1 << 5 //32
const EXTERNAL_SERVICE = 1 << 6   //64

type DBH struct {
	ctx context.Context
	db  *bun.DB
}

// Service belongs-to one host
type Service struct {
	bun.BaseModel `bun:"table:services"`
	ID            uint   `bun:"id,pk,autoincrement"`
	Uuid          string `bun:"uuid,notnull"`
	Name          string
	ServiceName   string `bun:",scanonly"`
	HostId        uint
	Host          Host `bun:"rel:belongs-to,join:host_id=id"`
}

// Host has-many services
// Eigentlich wollen wir das, aber das ORM macht da keinen JOIN sondern
// (`service`.`host_id`=77) OR (`service`.`host_id`=78) OR (`service`.`host_id`=79)
// das können wir bei Systemen mit 100k services nicht bringen
// Diese Query müsste er machen SELECT Hosts.id, Hosts.name, Hosts.uuid, Services.id, Services.name, Services.uuid FROM hosts AS Hosts LEFT JOIN services AS Services ON Services.host_id = Hosts.id
type Host struct {
	bun.BaseModel `bun:"table:hosts"`
	ID            uint   `bun:"id,pk,autoincrement"`
	Uuid          string `bun:"uuid,notnull"`
	Name          string
	// Services      []*Service `bun:"rel:has-many,join:id=host_id"`
}

type ExternalMonitoring struct {
	bun.BaseModel `bun:"table:external_monitorings"`
	ID            uint `bun:"id,pk,autoincrement"`
	Name          string
	Description   string
	ContainerId   uint
	SystemType    string
	JsonData      string
}

type HostScheduledDowntime struct {
	bun.BaseModel      `bun:"table:statusengine_host_scheduleddowntimes"`
	Hostname           string `bun:"hostname"`
	InternalDowntimeId int64  `bun:"internal_downtime_id"`
	ScheduledStartTime int64  `bun:"scheduled_start_time"`
	ScheduledEndTime   int64  `bun:"scheduled_end_time"`
	Comment            string `bun:"comment_data"`
	AuthorName         string `bun:"author_name"`
}

type ServiceScheduledDowntime struct {
	bun.BaseModel      `bun:"table:statusengine_service_scheduleddowntimes"`
	Hostname           string `bun:"hostname"`
	ServiceDescription string `bun:"service_description"`
	InternalDowntimeId int64  `bun:"internal_downtime_id"`
	ScheduledStartTime int64  `bun:"scheduled_start_time"`
	ScheduledEndTime   int64  `bun:"scheduled_end_time"`
	Comment            string `bun:"comment_data"`
	AuthorName         string `bun:"author_name"`
}

func readConfigurationIni(path string) (string, error) {
	cfg, err := ini.Load(path)
	if err != nil {
		return "", err
	}
	clientSec, err := cfg.GetSection("client")
	if err != nil {
		return "", err
	}
	hostKey, err := clientSec.GetKey("host")
	if err != nil {
		return "", err
	}
	hostname := hostKey.String()
	userKey, err := clientSec.GetKey("user")
	if err != nil {
		return "", err
	}
	username := userKey.String()

	passwordKey, err := clientSec.GetKey("password")
	if err != nil {
		return "", err
	}
	password := passwordKey.String()

	portKey, err := clientSec.GetKey("port")
	if err != nil {
		return "", err
	}
	port, err := portKey.Int()
	if err != nil {
		port = 3306
	}

	databaseKey, err := clientSec.GetKey("database")
	if err != nil {
		return "", err
	}
	database := databaseKey.String()

	if hostname == "localhost" {
		if _, err := os.Stat("/var/lib/mysql/mysql.sock"); err == nil {
			// RHEL
			return fmt.Sprintf("%s:%s@unix(%s)/%s?charset=utf8mb4&parseTime=True&loc=Local", username, password, "/var/lib/mysql/mysql.sock", database), nil
		}

		// Debian / Ubuntu
		return fmt.Sprintf("%s:%s@unix(%s)/%s?charset=utf8mb4&parseTime=True&loc=Local", username, password, "/run/mysqld/mysqld.sock", database), nil
	}
	return fmt.Sprintf("%s:%s@tcp(%s:%d)/%s?charset=utf8mb4&parseTime=True&loc=Local", username, password, hostname, port, database), nil
}

// New connects to the database
func NewDatabase(ctx context.Context, cfgPath string) (*DBH, error) {
	dsn, err := readConfigurationIni(cfgPath)
	if err != nil {
		return nil, err
	}
	return NewFromDsn(ctx, "mysql", dsn)
}

// NewFromDsn connects to the database
func NewFromDsn(ctx context.Context, string, dsn string) (*DBH, error) {
	sqldb, err := sql.Open("mysql", dsn)

	if err != nil {
		return nil, err
	}

	db := bun.NewDB(sqldb, mysqldialect.New())
	//db.AddQueryHook(bundebug.NewQueryHook(bundebug.WithVerbose(true)))

	dbh := &DBH{
		ctx: ctx,
		db:  db,
	}

	return dbh, nil
}

// Close closes the db connection
func (dbh *DBH) Close() {
	dbh.db.Close()
}

// FetchHosts returns all Hosts known to openITCOCKPIT
func (dbh *DBH) FetchHosts() ([]Host, error) {
	var hosts []Host

	err := dbh.db.NewSelect().
		Model(&hosts).
		Scan(dbh.ctx)

	if err != nil {
		return nil, err
	}

	return hosts, nil
}

// FetchServices returns all Services with the corosponding Host known to openITCOCKPIT
func (dbh *DBH) FetchServices() ([]Service, error) {
	var services []Service

	err := dbh.db.NewSelect().
		Model(&services).
		//Relation("Host").
		Column("service.id", "service.uuid", "service.name", "service.host_id").
		ColumnExpr("host.id AS host__id").
		ColumnExpr("host.uuid AS host__uuid").
		ColumnExpr("host.name AS host__name").
		ColumnExpr("IF((service.name IS NULL OR service.name=\"\"), servicetemplate.name, service.name) AS service_name").
		Join("INNER JOIN servicetemplates AS servicetemplate ON service.servicetemplate_id = servicetemplate.id").
		Join("INNER JOIN hosts AS host ON service.host_id = host.id").
		Where("service.service_type = ?", EXTERNAL_SERVICE).
		Scan(dbh.ctx)

	if err != nil {
		return nil, err
	}

	return services, nil
}

func (dbh *DBH) GetInternalDowntimeIdToCancelHostDowntime(downtime CancelDowntime) (int64, error) {

	var result HostScheduledDowntime

	err := dbh.db.NewSelect().
		Model(&result).
		Column("hostname", "internal_downtime_id", "scheduled_start_time", "scheduled_end_time", "comment_data", "author_name").
		Where("hostname = ?", downtime.HostUuid).
		Where("scheduled_start_time = ?", downtime.StartTime).
		Where("scheduled_end_time = ?", downtime.EndTime).
		Where("comment_data = ?", downtime.Comment).
		Limit(1).
		Scan(dbh.ctx)

	if err != nil {
		return 0, err
	}

	return result.InternalDowntimeId, nil

}

func (dbh *DBH) GetInternalDowntimeIdToCancelServiceDowntime(downtime CancelDowntime) (int64, error) {

	var result ServiceScheduledDowntime

	err := dbh.db.NewSelect().
		Model(&result).
		Column("hostname", "service_description", "internal_downtime_id", "scheduled_start_time", "scheduled_end_time", "comment_data", "author_name").
		Where("hostname = ?", downtime.HostUuid).
		Where("service_description = ?", downtime.ServiceUuid).
		Where("scheduled_start_time = ?", downtime.StartTime).
		Where("scheduled_end_time = ?", downtime.EndTime).
		Where("comment_data = ?", downtime.Comment).
		Limit(1).
		Scan(dbh.ctx)

	if err != nil {
		return 0, err
	}

	return result.InternalDowntimeId, nil

}

// GetExternalSystems returns all the configuration of all external monitoring systems defined in the external_monitorings database table
func (dbh *DBH) GetExternalSystems() ([]ExternalMonitoring, error) {
	var externalMonitorings []ExternalMonitoring

	err := dbh.db.NewSelect().
		Model(&externalMonitorings).
		Scan(dbh.ctx)

	if err != nil {
		return nil, err
	}

	return externalMonitorings, nil
}
