package oitc

import (
	"context"
	"fmt"
	"log"
	"sync"
	"time"
)

// The Database Worker is used to query the openITCOCKPIT MySQL database
// to cancel running downtimes
type DBWorker struct {
	wg  sync.WaitGroup
	mtx sync.Mutex

	shutdown          chan struct{}
	cancelDowntime    chan *CancelDowntime   // Reads downtimes to cancel from this channel
	rawCommandChannel chan *BrokerRawCommand // Send external commands to FIFO

	dbh *DBH
}

// run the event loop in a go func in background
func (w *DBWorker) run(parent context.Context) {
	ctx, cancel := context.WithCancel(parent)

	defer func() {
		cancel()
		close(w.cancelDowntime)
		w.wg.Done()
	}()

	for {
		select {
		case <-ctx.Done():
			return
		case _, more := <-w.shutdown:
			if !more {
				// Channel c.shutdown got closed - close database worker
				return
			}
		case downtime := <-w.cancelDowntime:
			// Query the database if a known downtime exists for the provided parameters
			// and try to cancel it if possible

			if downtime.ServiceUuid == "" {
				// Try to find host downtime
				w.mtx.Lock()
				internalDowntimeId, err := w.dbh.GetInternalDowntimeIdToCancelHostDowntime(*downtime)
				w.mtx.Unlock()
				if err != nil {
					log.Println("Error while looking up internal downtime id to cancel host downtime. Error: ", err)
				}

				if internalDowntimeId > 0 {
					// Delete Host Downtime
					cmd := fmt.Sprintf(
						"[%d] DEL_HOST_DOWNTIME;%d",
						time.Now().Unix(),
						internalDowntimeId,
					)

					//  Send acknowledgement as external command to FIFO
					w.rawCommandChannel <- &BrokerRawCommand{
						Command: "raw",
						Data:    cmd,
					}
				}
			} else {
				// Try to find service downtime
				w.mtx.Lock()
				internalDowntimeId, err := w.dbh.GetInternalDowntimeIdToCancelServiceDowntime(*downtime)
				w.mtx.Unlock()
				if err != nil {
					log.Println("Error while looking up internal downtime id to cancel host downtime. Error: ", err)
				}

				if internalDowntimeId > 0 {
					// Delete Service Downtime
					cmd := fmt.Sprintf(
						"[%d] DEL_SVC_DOWNTIME;%d",
						time.Now().Unix(),
						internalDowntimeId,
					)

					//  Send acknowledgement as external command to FIFO
					w.rawCommandChannel <- &BrokerRawCommand{
						Command: "raw",
						Data:    cmd,
					}
				}
			}

		}
	}
}

// Start the event loop and returns if successful
func (w *DBWorker) Start(ctx context.Context, dbh *DBH, raw chan *BrokerRawCommand) {
	w.wg.Add(1)

	// Signal this channel to stop the database worker
	w.shutdown = make(chan struct{})

	// Channel to FIFO to submit external commands
	w.rawCommandChannel = raw

	// Downtimes that need to be canceled will be send to this channel by the poller threads
	w.cancelDowntime = make(chan *CancelDowntime)

	w.dbh = dbh

	// Start new database worker in separate go func (thread) and return from this function
	go w.run(ctx)
}

// Stop should be called before programs exit
func (w *DBWorker) Stop() {
	close(w.shutdown)
	w.wg.Wait()
}

// DirectCancelDowntime returns the channel where Icinga (or other sources) can cancel running downtimes
func (w *DBWorker) DirectCancelDowntime() chan *CancelDowntime {
	return w.cancelDowntime
}
