package oitc

import "fmt"

type OpenITCOCKPITMapping struct {
	hostCache    map[string]string
	serviceCache map[string]map[string]string
	dbh          *DBH // Database Handler for MySQL
}

func (o *OpenITCOCKPITMapping) Reload() error {
	// Query hosts from database
	hosts, err := o.dbh.FetchHosts()
	if err != nil {
		return err
	}

	hostMapping := make(map[string]string)
	for _, host := range hosts {
		// Add host to mapping
		hostMapping[host.Name] = host.Uuid
	}

	// Query services from database
	services, err := o.dbh.FetchServices()
	if err != nil {
		return err
	}

	serviceMapping := make(map[string]map[string]string)
	for _, service := range services {
		// Check that host UUID exists
		if _, ok := serviceMapping[service.Host.Uuid]; !ok {
			// Host UUID not found, create nested map
			serviceMapping[service.Host.Uuid] = make(map[string]string)
		}

		// Add service to mapping
		serviceMapping[service.Host.Uuid][service.ServiceName] = service.Uuid
	}

	// We create the new mapping in a local variable and overwrite the struct when everything is done in the local variable
	// This will hopefully prevent us from adding a mutex
	o.hostCache = make(map[string]string)
	o.serviceCache = make(map[string]map[string]string)
	o.hostCache = hostMapping
	o.serviceCache = serviceMapping
	return nil
}

func NewOitcMapping(dbh *DBH) *OpenITCOCKPITMapping {
	mapping := &OpenITCOCKPITMapping{
		hostCache:    make(map[string]string),
		serviceCache: make(map[string]map[string]string),
		dbh:          dbh,
	}

	return mapping
}

func (o *OpenITCOCKPITMapping) GetHostUuidByHostname(hostname string) (string, error) {
	if val, ok := o.hostCache[hostname]; ok {
		if val == "" {
			return "", fmt.Errorf("UUID found for host %v is an empty string. Impossible", hostname)
		}

		return val, nil
	}

	return "", fmt.Errorf("no UUID found for host %v", hostname)
}

func (o *OpenITCOCKPITMapping) GetServiceUuidByHostUuidAndServiceDescription(hostUuid string, serviceDescription string) (string, error) {
	if _, ok := o.serviceCache[hostUuid]; ok {
		// Host UUID exists in cache

		if val, sOk := o.serviceCache[hostUuid][serviceDescription]; sOk {
			// Service description found in cache - return the service UUID
			return val, nil
		}
		return "", fmt.Errorf("no UUID found for host uuid/service %v/%v", hostUuid, serviceDescription)
	}

	return "", fmt.Errorf("no record for host UUID %v in service cache", hostUuid)
}
